import os
import concurrent.futures
import hashlib
import pathlib
from typing import Tuple, Optional, List, Union
from tqdm import tqdm
import numpy as np
import vdf
import platform
from string import ascii_uppercase

# not expecting BIOS files over 15mb
MAX_BIOS_BYTES = 15728640


def hash_file(file_path: pathlib.Path) -> str:
    """
    This function calculates the MD5 hash of a file.

    :param file_path:
    :return:
    """

    file_bytes = file_path.read_bytes()
    file_hash = hashlib.md5(file_bytes)
    return file_hash.hexdigest()


def recursive_hash(
    directory: pathlib.Path, glob: str = "*"
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Calculate the MD5 hash for all files that match the glob pattern (recursively).

    :param directory: Starting directory for the glob pattern matching
    :param glob: The glob pattern to match files. Defaults to "*".
    :return: An array with the file_paths to selected files and an array with the corresponding MD5 hashes
    """

    file_paths = list(directory.rglob(pattern=glob))
    file_paths = [
        file_path
        for file_path in file_paths
        if file_path.exists()
        and file_path.stat().st_size <= MAX_BIOS_BYTES
        and file_path.is_file()
    ]

    file_hashes = []
    with concurrent.futures.ThreadPoolExecutor() as executor:
        for file_hash in tqdm(
            executor.map(hash_file, file_paths),
            total=len(file_paths),
            desc="Hashing files",
        ):
            file_hashes.append(file_hash)
    return np.array(file_paths), np.array(file_hashes)


def match_arrays(
    array_a: np.ndarray, array_b: np.ndarray
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Element-wise array comparison, returns unique values and matching indices per input array

    :param array_a: The first array to compare.
    :param array_b: The second array to compare.
    :return: A tuple of three numpy arrays: unique matching values, indices of the matching values in
    array_a and indices of the matching values in array_b.
    """

    # expecting 1D arrays
    if np.sum([len(array_a.shape), len(array_b.shape)]) > 2:
        raise ValueError("input arrays need to be one-dimensional, exiting..")

    comparisons = np.equal(array_a.reshape(1, -1), array_b.reshape(-1, 1))

    indices_b, indices_a = np.where(comparisons)
    matching_values = np.unique(array_a[indices_a])

    return matching_values, indices_a, indices_b


def list_steam_libraries(vdf_path: pathlib.Path) -> List[pathlib.Path]:
    """
    Getting paths to steam libraries from libraryfolders.vdf (Valve Data File)

    :param vdf_path: The path to libraryfolders.vdf
    :return: List of paths to Steam Library locations
    """

    library_paths = []
    with open(vdf_path, "r", encoding="utf-8") as src:
        library_info = vdf.parse(src)
        for key in library_info["libraryfolders"].keys():
            library_path = pathlib.Path(library_info["libraryfolders"][key]["path"])
            if library_path.exists():
                library_paths.append(library_path)

    return library_paths


def find_retroarch() -> Optional[pathlib.Path]:
    """
    Find the path to the RetroArch installation in the system.

    :return: The path to the RetroArch installation if found, None otherwise.
    """

    paths_to_check = []
    system = platform.system()

    if system == "Windows":
        system_glob = "RetroArch*/system"
        # Non-Steam
        drives = [
            pathlib.Path(f"{drive}:").resolve()
            for drive in ascii_uppercase
            if pathlib.Path(f"{drive}:").exists()
        ]
        for drive in drives:
            paths_to_check.append(drive)

        env_vars = ["PROGRAMFILES(X86)", "PROGRAMFILES"]
        for env_var in env_vars:
            env_var = os.environ.get(env_var)
            if not env_var:
                continue
            env_path = pathlib.Path(env_var)

            if env_path.exists():
                paths_to_check.append(env_path)

                # adding steam libraries
                vdf_path = env_path / pathlib.Path(
                    "Steam", "steamapps", "libraryfolders.vdf"
                )
                if vdf_path.exists():
                    for library_path in list_steam_libraries(vdf_path=vdf_path):
                        paths_to_check.append(
                            library_path / pathlib.Path("steamapps/common")
                        )

    elif system == "Linux":
        system_glob = "retroarch/system"
        home = pathlib.Path.home()
        paths_to_check.append(home / pathlib.Path(".config"))
        vdf_path = home / pathlib.Path(
            ".local/share/Steam/steamapps/libraryfolders.vdf"
        )
        if vdf_path.exists():
            for library_path in list_steam_libraries(vdf_path=vdf_path):
                paths_to_check.append(library_path / pathlib.Path("steamapps/common"))

    # requires more testing on actual metal
    elif system == "Darwin":
        system_glob = "RetroArch.app/Contents/Resources/system"
        home = pathlib.Path.home()
        paths_to_check.append(home / pathlib.Path("Library/Application Support"))
        vdf_path = home / pathlib.Path(
            "Library/Application Support/Steam/steamapps/libraryfolders.vdf"
        )
        if vdf_path.exists():
            for library_path in list_steam_libraries(vdf_path=vdf_path):
                paths_to_check.append(library_path / pathlib.Path("steamapps/common"))

    # checking for retroarch/system (one level down)
    for path_to_check in paths_to_check:
        # glob is needed for inconsistent parent naming (e.g. RetroArch-Win32, RetroArch-Win64, retroarch)
        for path in path_to_check.glob(system_glob):
            return path
    return None



import shutil
import pathlib
import numpy as np
from gooey import Gooey, GooeyParser

from config import SYSTEMS as system_df
from config import RETROARCH_PATH
from typing import List, Optional
from libretro_finder.utils import match_arrays, recursive_hash


def organize(search_dir: pathlib.Path, output_dir: pathlib.Path) -> None:
    """
    Non-destructive function that finds, copies and refactors files to the format expected by
    libretro (and its cores). This is useful if you source your BIOS files from many different
    places and have them saved them under different names (often with duplicates).

    :param search_dir: starting location of recursive search
    :param output_dir: path to output directory (will be created if it doesn't exist)
    :return:
    """

    # Indexing files to be checked for matching MD5 checksums
    output_dir.mkdir(parents=True, exist_ok=True)
    file_paths, file_hashes = recursive_hash(directory=search_dir)

    # Element-wise matching of files against libretro's files
    matching_values, file_indices, system_indices = match_arrays(
        array_a=file_hashes, array_b=np.array(system_df["md5"].values)
    )

    if not np.size(matching_values) > 0:
        print("No matching BIOS files were found, exiting..")
        return

    # If a single core has multiple matches, we just pick the first one
    _, indices = np.unique(system_indices, return_index=True)
    srcs = file_paths[file_indices[indices]]
    hashes = file_hashes[file_indices[indices]]
    system_subset = system_df.loc[system_indices[indices]]

    # np.unique and indexing doesn't merit a dedicated function but it should still be tested
    assert np.array_equal(np.array(system_subset["md5"].values), hashes)
    assert system_subset["name"].size == system_subset["name"].unique().size

    # printing matches per system
    matches = system_subset.groupby("system").count()
    print(
        f"{matches['name'].sum()} matching BIOS files were found for {matches.shape[0]} "
        "unique systems:"
    )
    for name, row in matches.iterrows():
        print(f"\t{name} ({row['name']})")

    # copying matching files to output_dir using structure specified by libretro
    dsts = system_subset["name"].values

    # checking whether our input and output paths are of equal length
    assert len(srcs) == len(dsts)

    for i in range(srcs.size):
        dst = output_dir / dsts[i]
        parent = dst.parent
        if dst.exists() or srcs[i] == dst:
            continue
        if parent != output_dir:
            parent.mkdir(parents=True, exist_ok=True)

        shutil.copy(src=srcs[i], dst=dst)


@Gooey(program_name="LibretroFinder", default_size=(610, 530), required_cols=1)
def main(argv: Optional[List[str]] = None) -> None:
    """
    A simple command line utility that finds and prepares your BIOS files for all documented
    RetroArch cores. If called without any arguments, a simple graphical user interface with
    the same functionality will be started (courtesy of Gooey).
    """

    parser = GooeyParser(
        description="Locate and prepare your BIOS files for libretro.",
    )
    parser.add_argument(
        "Search directory",
        help="Where to look for BIOS files",
        type=pathlib.Path,
        widget="DirChooser",
    )
    parser.add_argument(
        "Output directory",
        help="Where to output refactored BIOS files (defaults to ./retroarch/system)",
        type=pathlib.Path,
        widget="DirChooser",
        default=str(RETROARCH_PATH) if RETROARCH_PATH else None,
    )
    args = vars(parser.parse_args(argv))

    search_directory = args["Search directory"]
    output_directory = args["Output directory"]

    if not search_directory.exists():
        raise FileNotFoundError("Search directory does not exist..")
    elif not search_directory.is_dir():
        raise NotADirectoryError("Search directory needs to be a directory..")

    organize(search_dir=search_directory, output_dir=output_directory)


if __name__ == "__main__":
    main()
